# SPDX-License-Identifier: GPL-2.0-or-later
# pylint: disable=R0801
"""
repository helpers
"""
import subprocess
import re
from pathlib import Path
from typing import List, Optional, Union
from bs4 import BeautifulSoup  # type: ignore
from .repo import Repo, _logger, shortest_path


class SvnInfo:
    # pylint: disable=too-many-instance-attributes, too-few-public-methods
    """
    container class obtaining and holding svn repository information

    Attributes
    ----------
    kind : str
      node kind of TARGET
    url : str
        URL of TARGET in the repository
    relative_url : str
        repository-relative URL of TARGET
    repos_root_url : str
        root URL of repository
    repos_uuid : str
        UUID of repository
    revision : str
        specified or implied revision
    last_changed_revision : str
        last change of TARGET at or before revision
    last_changed_date : str
        date of last_changed_revision
    last_changed_author : str
        author of last_changed_revision
    wc_root : str
        root of TARGETs working copy
    schedule : str
        normal,add,delete,replace
    depth : str
        checkout depth of TARGET in WC
    """

    def __init__(self, uri: Union[str, Path]):
        self.update(uri)

    def update(self, uri: Optional[Union[str, Path]] = None):
        """
        update svn info using data scraped from uri, which can be both a
        local checkout or a remote repository
        """
        if uri is not None:
            self.uri = uri
        info = BeautifulSoup(
            subprocess.run(
                ["svn", "info", "--xml", self.uri], capture_output=True, check=False
            ).stdout.decode(),
            "xml",
        ).find("entry")
        repo = info.find("repository")
        workingcopy = info.find("wc-info")
        commit = info.find("commit")

        self.kind = info["kind"]
        self.url = info.find("url").text
        self.relative_url = info.find("relative-url").text
        self.revision = info["revision"]
        if repo is not None:
            self.repos_root_url = repo.find("root").text
            self.repos_uuid = repo.find("uuid").text
        else:
            self.repos_root_url = ""
            self.repos_uuid = ""

        if commit is not None:
            self.last_changed_revision = commit["revision"]
            self.last_changed_date = commit.find("date").text
            self.last_changed_author = commit.find("author").text
        else:
            self.last_changed_revision = ""
            self.last_changed_date = ""
            self.last_changed_author = ""

        if workingcopy is not None:
            self.wc_root = workingcopy.find("wcroot-abspath").text
            self.schedule = workingcopy.find("schedule").text
            self.depth = workingcopy.find("depth").text
        else:
            self.wc_root = ""
            self.schedule = ""
            self.depth = ""


class SvnRepo(Repo):
    """
    svn repository helper implementation
    """

    vcs = "svn"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        url: Optional[str],
        path: Path,
        target_commit: Optional[str] = None,
        tag: Optional[str] = None,
        branch: Optional[str] = None,
        remote_path: Optional[str] = None,
        options: Optional[List[str]] = None,
    ):
        rurl = url
        rpath = remote_path
        infotarget = ""
        if (path / ".svn").exists():
            infotarget = str(path)
        elif url is not None:
            infotarget = url
        self.info = SvnInfo(infotarget)
        if url is not None:
            rurl = self.info.repos_root_url
            toks = self.info.relative_url[2:].split("/")
            if toks[0] == "tags" and len(toks) > 1:
                tag = toks[1]
                rpath = "/".join(toks[2:])
            elif toks[0] == "branches" and len(toks) > 1:
                branch = toks[1]
                rpath = "/".join(toks[2:])
            elif toks[0] == "trunk":
                rpath = "/".join(toks[1:])
            if rpath == "":
                rpath = remote_path
        super().__init__(
            rurl,
            path,
            target_commit=target_commit,
            tag=tag,
            branch=branch,
            remote_path=rpath,
            options=options,
        )

    @property
    def vcs_dir(self):
        return self.path / ".svn"

    def get_url(self) -> Optional[str]:
        return (
            self.execute(
                "info --show-item url".split(),
                capture_output=True,
                check=False,
            )
            .stdout.decode()
            .strip()
        )

    def add(self, file: Path):
        _logger.info("Adding %s", file)
        self.execute(f"add --parents {file}".split(), check=False, capture_output=True)

    def remove(self, file: Path):
        _logger.info("Removing %s", file)
        self.execute(f"rm --force {file}".split(), check=False, capture_output=True)

    def checkout(self):
        cmd = ["co"]
        if self.target_commit is not None:
            cmd.append("-r" + self.target_commit)
        cmd.extend(self.options)

        url = self.url
        if self.tag is not None:
            url += f"/tags/{self.tag}"
        elif self.branch is not None:
            if self.branch == "trunk":
                url += "/trunk"
            else:
                url += f"/branches/{self.branch}"
        else:
            url += "/trunk"
        if self.remote_path is not None:
            url += f"/{self.remote_path}"
        cmd.append(url)

        cmd.append(self.path)
        _logger.info("Checking out %s into %s", url, shortest_path(self.path))
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.execute(cmd, cwd=self.path.parent)
        self.info.update(self.path)

    def update(self):
        _logger.info("Updating %s", shortest_path(self.path))
        cmd = ["update"]
        if self.target_commit is not None:
            cmd.append(f"-r{self.target_commit}")
        self.execute(cmd)
        self.info.update(self.path)

    def status(self):
        stat = (
            self.info.relative_url
            + "\n"
            + self.execute("status", capture_output=True).stdout.decode()
        )

        dirty = ""
        if self.is_dirty():
            dirty = "[D] "

        _logger.info("%s: %s%s\n", shortest_path(self.path), dirty, stat.strip())

    def has_externals(self, relpath: str = ".", recursive: bool = True):
        if not (self.path / relpath).exists():
            return False
        if recursive:
            cmd = "pget -R".split()
        else:
            cmd = "pget".split()
        cmd.append("svn:externals")
        cmd.append(relpath)

        if recursive:
            return (
                self.execute(cmd, capture_output=True).stdout.decode().strip() != ". -"
            )

        return self.execute(cmd, check=False, capture_output=True).returncode == 0

    def is_dirty(self, relpath: str = "."):
        out = self.execute(
            f"st --xml --depth immediates {relpath}".split(),
            capture_output=True,
        ).stdout
        soup = BeautifulSoup(out, "xml")
        regex = re.compile("unversioned|modified|added|deleted")
        return bool(soup.find_all("wc-status", attrs={"item": regex}))

    def list_externals(self, relpath: str = ".") -> List[Repo]:
        return_list: List[Repo] = []

        if not (self.path / relpath).exists():
            return return_list

        try:
            out = self.execute(
                f"pget svn:externals {relpath}".split(), capture_output=True
            )

            for line in out.stdout.decode().split("\n"):
                if len(line.strip()) > 0:
                    _logger.debug("processing external: '%s'", line)
                    ext = line.split()
                    options = []
                    target_commit = None
                    for opt in ext[1:-1]:
                        if re.match("^-r[0-9]+$", opt):
                            target_commit = opt[2:]
                        else:
                            options.append(opt)

                    return_list.append(
                        SvnRepo(
                            ext[-1],
                            self.path / relpath / ext[0],
                            target_commit=target_commit,
                            options=options,
                        )
                    )
        except subprocess.CalledProcessError:
            pass

        return return_list

    def list_files(self, printout: bool = True):
        return self._return_local_files(
            self.execute("ls -R".split(), capture_output=True).stdout.decode(), printout
        )

    def list_folders(self, printout: bool = True):
        return self._return_local_folders(
            self.execute("ls -R".split(), capture_output=True).stdout.decode(), printout
        )

    def rm_externals(self, relpath: str = "."):
        _logger.info("Delete svn:externals property from %s", relpath)
        self.execute(f"pdel svn:externals {relpath}".split())

    def reset_externals(self, relpath: str = "."):
        _logger.info("Restore svn:externals property for %s", relpath)
        self.execute(f"revert --depth empty {relpath}".split())

    def add_ignores(self, *patterns: Path):
        ignores_map = {}
        for pattern in patterns:
            relpath = pattern.relative_to(self.path).parent
            if relpath not in ignores_map:
                ignores_map[relpath] = (
                    False,
                    [
                        i.strip()
                        for i in self.execute(
                            f"pget svn:ignore {relpath}".split(),
                            check=False,
                            capture_output=True,
                        )
                        .stdout.decode()
                        .splitlines()
                    ],
                )
            if pattern.name in ignores_map[relpath][1]:
                _logger.info(
                    "Ignore rule %s already exists in %s", pattern.name, relpath
                )
            else:
                _logger.info("Append ignore rule %s to %s", pattern.name, relpath)
                ignores_map[relpath] = (True, [*ignores_map[relpath][1], pattern.name])

        for relpath, ignores in ignores_map.items():
            if ignores[0]:
                self.add(relpath)

                cmd = "pset svn:ignore".split()
                cmd.append("\n".join(sorted(ignores[1])))
                cmd.append(str(relpath))
                self.execute(cmd)

    def del_ignores(self, *patterns: Path):
        ignores_map = {}
        for pattern in patterns:
            relpath = pattern.relative_to(self.path).parent
            if relpath not in ignores_map:
                ignores_map[relpath] = (
                    False,
                    [
                        i.strip()
                        for i in self.execute(
                            f"pget svn:ignore {relpath}".split(),
                            check=False,
                            capture_output=True,
                        )
                        .stdout.decode()
                        .splitlines()
                    ],
                )
            if pattern.name in ignores_map[relpath][1]:
                _logger.info("Delete ignore rule %s from %s", pattern, relpath)
                ignores_map[relpath] = (
                    True,
                    [i for i in ignores_map[relpath][1] if i != pattern.name],
                )

        for relpath, ignores in ignores_map.items():
            if ignores[0]:
                cmd = "pset svn:ignore".split()
                cmd.append("\n".join(sorted(ignores[1])))
                cmd.append(str(relpath))
                self.execute(cmd)

    def commit(self, message: str, files: List[Path]):
        raise NotImplementedError()

    @staticmethod
    def check_remote(url):
        cmd = f"svn info {url}".split()
        # pylint: disable=subprocess-run-check
        return subprocess.run(cmd, capture_output=True).returncode == 0



# SPDX-License-Identifier: GPL-2.0-or-later
# pylint: disable=R0801
"""
repository helpers
"""
import logging
import os
import subprocess
from pathlib import Path
from typing import Union, List, Optional
from abc import ABC, abstractmethod

_logger = logging.getLogger(__name__)


def shortest_path(path: Path) -> Path:
    """
    return the shortest representation of input path:
    - a relative path if it is inside current directory
    - absolute path otherwise
    """
    try:
        return path.resolve().relative_to(Path(".").resolve())
    except ValueError:
        return path.resolve()


class Repo(ABC):
    # pylint: disable=too-many-public-methods
    """
    base abstract class for repository helper implementations, defines the
    required public methods
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        url: Optional[str],
        path: Path,
        target_commit: Optional[str] = None,
        tag: Optional[str] = None,
        branch: Optional[str] = None,
        remote_path: Optional[str] = None,
        options: Optional[List[str]] = None,
    ):
        self.url = url
        self.path = path
        self.target_commit = target_commit
        self.tag = tag
        self.branch = branch
        self.remote_path = remote_path
        self.options = []
        if isinstance(options, list):
            self.options = options

        if self.url is None:
            self.url = self.get_url()

    @property
    @abstractmethod
    def vcs(self) -> str:
        """
        return vcs type string
        """

    global_options: List[str] = []
    """
    optional list of cmdline options to be used for each command
    """

    @property
    @abstractmethod
    def vcs_dir(self) -> Path:
        """
        return resolved vcs private directory
        """

    def __str__(self):
        res = f"{shortest_path(self.path)} ("
        if self.remote_path is not None:
            res += f"{self.remote_path} "
        res += f"{self.vcs}: {self.url} -"
        if self.tag is not None:
            res += f" T'{self.tag}'"
        elif self.branch is not None:
            res += f" b'{self.branch}'"
        if self.target_commit is not None:
            res += f" r{self.target_commit}"
        res += f" {self.options})"
        return res

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            self.url == other.url
            and self.path == other.path
            and self.target_commit == other.target_commit
            and self.tag == other.tag
            and self.branch == other.branch
            and self.remote_path == other.remote_path
            and self.options == other.options
        )

    @abstractmethod
    def get_url(self) -> Optional[str]:
        """
        Retrieve URL from local repo
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def add(self, file: Path):
        """
        Add a file/directory to current local repo,
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def remove(self, file: Path):
        """
        Remove a file/directory from current local repo,
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def checkout(self):
        """
        perform a fresh checkout of the remote repository into local path,
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def update(self):
        """
        update the local copy of the repository pulling remote changes
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def status(self):
        """
        print out the local repository status message
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def has_externals(self, relpath: str = ".", recursive: bool = True):
        """
        return True if local repository has implementation defined externals
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def is_dirty(self, relpath: str = "."):
        """
        return True if repository has local modification
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def list_externals(self, relpath: str = "."):
        """
        return a list of external subprojects defined in this repository as a list of Repo
        instances
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def list_files(self, printout: bool = True):
        """
        return the list of files handled by this repository prefixed by the repo path
        if printout is true also print to console
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def list_folders(self, printout: bool = True):
        """
        return list of folder commited in this repository
        if printout is true also print to console
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def rm_externals(self, relpath: str = "."):
        """
        remove external subprojects defined in this repository
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def reset_externals(self, relpath: str = "."):
        """
        reset externals to remote original state
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def add_ignores(self, *patterns: Path):
        """
        add ignore patterns to local repository ignores list
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def del_ignores(self, *patterns: Path):
        """
        delete ignore patterns from local repository ignores list
        MUST be implemented by actual Repo helper implementation
        """

    @abstractmethod
    def commit(self, message: str, files: List[Path]):
        """
        commit files with defined message
        MUST be implemented by actual Repo helper implementation
        """

    def execute(self, command_and_args: Union[str, List[str]], **subprocess_kwargs):
        """
        execute a command in the local repository path
        """
        if isinstance(command_and_args, str):
            command_and_args = [command_and_args]
        cmd = [self.vcs]
        cmd.extend(self.global_options)
        cmd.extend([str(arg) for arg in command_and_args])
        if "cwd" not in subprocess_kwargs:
            subprocess_kwargs["cwd"] = self.path
        if "check" not in subprocess_kwargs:
            subprocess_kwargs["check"] = True
        _logger.debug(
            "executing command: '%s', (env: %s)", " ".join(cmd), subprocess_kwargs
        )
        # pylint: disable=subprocess-run-check
        return subprocess.run(cmd, **subprocess_kwargs)

    def exists(self):
        """
        return True if local repository exists
        """
        return self.path.exists()

    def _return_local_files(self, output_list: str, printout: bool):
        return_list = [str(self.path / line) for line in output_list.split("\n")]

        if printout:
            _logger.info(
                "Listing %s files:\n%s",
                shortest_path(self.path),
                "\n".join(return_list),
            )
        return return_list

    def _return_local_folders(self, output_list: str, printout: bool):
        return_list = []
        for line in output_list.split("\n"):
            if os.path.isdir(self.path / line):
                return_list.append(str(self.path / line))

        if printout:
            _logger.info(
                "Listing %s folders:\n%s",
                shortest_path(self.path),
                "\n".join(return_list),
            )
        return return_list

    @staticmethod
    @abstractmethod
    def check_remote(url: str):
        """
        return True if remote url is a valid repository,
        MUST be implemented by actual Repo helper implementation
        """
